(*  Title:      Nitpick/examples/InductNits.thy
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2009

Examples featuring Nitpick applied to (co)inductive definitions.
*)

header {* Examples Featuring Nitpick Applied to (Co)inductive Definitions *}

theory InductNits
imports Main "../Nitpick"
begin

nitpick_params [verbose, show_all]

inductive p1 :: "nat \<Rightarrow> bool" where
"p1 0" |
"p1 n \<Longrightarrow> p1 (n + 2)"

coinductive q1 :: "nat \<Rightarrow> bool" where
"q1 0" |
"q1 n \<Longrightarrow> q1 (n + 2)"

lemma "p1 = q1"
nitpick [expect = none]
nitpick [inductive_mood = optimistic, expect = none]
nitpick [inductive_mood = pessimistic, expect = none]
oops

lemma "p1 \<noteq> q1"
nitpick [expect = potential]
nitpick [inductive_mood = optimistic, expect = potential]
nitpick [inductive_mood = pessimistic, expect = potential]
oops

lemma "p1 (n - 2) \<Longrightarrow> p1 n"
nitpick [expect = genuine]
nitpick [inductive_mood = pessimistic, expect = genuine]
oops

lemma "q1 (n - 2) \<Longrightarrow> q1 n"
nitpick [expect = genuine]
nitpick [inductive_mood = pessimistic, expect = genuine]
oops

inductive p2 :: "nat \<Rightarrow> bool" where
"p2 n \<Longrightarrow> p2 n"

coinductive q2 :: "nat \<Rightarrow> bool" where
"q2 n \<Longrightarrow> q2 n"

lemma "p2 = {}"
nitpick [expect = none]
sorry

lemma "q2 = {}"
nitpick [expect = genuine]
nitpick [inductive_mood = optimistic, expect = likely_genuine]
oops

lemma "p2 = UNIV"
nitpick [expect = genuine]
oops

lemma "q2 = UNIV"
nitpick [expect = none]
nitpick [inductive_mood = optimistic, expect = likely_genuine]
sorry

lemma "p2 = q2"
nitpick [expect = genuine]
oops

lemma "p2 n"
nitpick [expect = genuine]
nitpick [special_depth = -1, expect = genuine]
oops

lemma "q2 n"
nitpick [expect = none]
sorry

lemma "\<not> p2 n"
nitpick [expect = none]
sorry

lemma "\<not> q2 n"
nitpick [expect = genuine]
nitpick [special_depth = -1, expect = genuine]
oops

inductive p3 and p4 where
"p3 0" |
"p3 n \<Longrightarrow> p4 (Suc n)" |
"p4 n \<Longrightarrow> p3 (Suc n)"

coinductive q3 and q4 where
"q3 0" |
"q3 n \<Longrightarrow> q4 (Suc n)" |
"q4 n \<Longrightarrow> q3 (Suc n)"

lemma "p3 = q3"
nitpick [expect = none]
nitpick [inductive_mood = pessimistic, expect = none]
sorry

lemma "p4 = q4"
nitpick [expect = none]
nitpick [inductive_mood = pessimistic, expect = none]
sorry

lemma "p3 = UNIV - p4"
nitpick [expect = none]
nitpick [inductive_mood = pessimistic, expect = none]
sorry

lemma "q3 = UNIV - q4"
nitpick [expect = none]
nitpick [inductive_mood = pessimistic, expect = none]
sorry

lemma "p3 \<inter> q4 \<noteq> {}"
nitpick [expect = potential]
nitpick [inductive_mood = pessimistic, expect = potential]
sorry

lemma "q3 \<inter> p4 \<noteq> {}"
nitpick [expect = potential]
nitpick [inductive_mood = pessimistic, expect = potential]
sorry

lemma "p3 \<union> q4 \<noteq> UNIV"
nitpick [expect = potential]
nitpick [inductive_mood = pessimistic, expect = potential]
sorry

lemma "q3 \<union> p4 \<noteq> UNIV"
nitpick [expect = potential]
nitpick [inductive_mood = pessimistic, expect = potential]
sorry

end
